#include "utils.h"
#include <stdint.h>

void
dotproduct_omp_critical(double * const sumref, 
	   double const * const A, 
	   double const * const B, 
	   const unsigned long  N);

void
dotproduct_omp_atomic(double * const sumref, 
	   double const * const A, 
	   double const * const B, 
	   const unsigned long  N);

void
dotproduct_omp(double * const sumref, 
	   double const * const A, 
	   double const * const B, 
	   const unsigned long  N);


#ifndef NUMTHREADS 
#define NUMTHREADS 4UL
#endif

#ifndef N
#define N 100UL
#endif

void
dotproduct_seq(double *sumref, double *A, double *B, const size_t UPPER)
{
    double sum = 0.0;
    for (uint64_t i = 0; i < UPPER; ++i)
        sum += A[i] * B[i];
    *sumref = sum;
}

int
main(void)
{
    double *A         = NULL, 
           *B         = NULL;
    long    n         = N;
    double  sum       = 0.0;

    A = smalloc(sizeof(double) * N);
    B = smalloc(sizeof(double) * N);

    for (long i = 0; i < n; ++i) {
        A[i] = i;
        B[i] = n-i;
    }

    dotproduct_seq(&sum,A,B,n);
    printf("Sequential dot product                     = %.2f\n", sum);

    sum = 0.0;
    dotproduct_omp(&sum,A,B,n);
    printf("Parallel (OpenMP) dot product -- reduction = %.2f\n", sum);

    sum = 0.0;
    dotproduct_omp_critical(&sum,A,B,n);
    printf("Parallel (OpenMP) dot product -- critical  = %.2f\n", sum);

    sum = 0.0;
    dotproduct_omp_atomic(&sum,A,B,n);
    printf("Parallel (OpenMP) dot product -- atomic    = %.2f\n", sum);

    free(A);
    free(B);

    return 0;
}
