#include "dotproduct.h"


void *
dotproduct_task(void *arg)
{
    dotproduct_t *dp  = (dotproduct_t *) arg;
    double        sum = 0.0;

    uint64_t lo = dp->id     * dp->N/dp->nthreads;
    uint64_t hi = (dp->id+1) * dp->N/dp->nthreads;
  
    for (uint64_t i = lo; i < hi; ++i) 
        sum += dp->A[i] * dp->B[i];

    spthread_mutex_lock(dp->lock);
    *(dp->sum) += sum;
    spthread_mutex_unlock(dp->lock);

    return NULL; /* no return state to give here */
}

void 
dotproduct_pthread(double *sumref, double * A, double * B, 
                   const uint64_t  N, const uint64_t  n_threads)
{
    dotproduct_t    *dp      = smalloc(sizeof(dotproduct_t) * n_threads);
    pthread_t       *threads = smalloc(sizeof(pthread_t)    * n_threads);
    pthread_mutex_t  lock    = PTHREAD_MUTEX_INITIALIZER;
    double           sum     = 0.0;

    pthread_attr_t attr;
    pthread_attr_init(&attr);
    /* pedantic: default value is already PTHREAD_CREATE_JOINABLE */
    pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE); 

    for (uint64_t i = 0; i < n_threads; ++i) {
        dp[i] = (dotproduct_t){ 
            .A=A, .B=B, .sum=&sum, .N=N, .id=i, .lock=&lock, .nthreads=n_threads 
        };
        spthread_create(&threads[i],&attr, dotproduct_task, &dp[i]);
    }
    for (uint64_t i = 0; i < n_threads; ++i)
        spthread_join(threads[i],NULL);

    *sumref = sum;
    free(threads);
    free(dp);
}

