#ifndef DOTPRODUCT_H_GUARD
#define DOTPRODUCT_H_GUARD

#include "utils.h"
#include "pthread_wrappers.h"

typedef struct dotproduct_atomic_s {
    double          *A, *B; 
    volatile double *sum;
    uint64_t         N, id, nthreads;
//  volatile uint64_t *synchronized;
} dotproduct_atomic_t;

typedef struct dotproduct_s {
    pthread_mutex_t *lock;
    double          *A, *B, *sum;
    uint64_t         N, id, nthreads;
} dotproduct_t;

static inline double atomic_add_dbl(volatile double* ptr, double value) {
    union { double dbl; uint64_t ui64; } newval, oldval;
    do {
        oldval.dbl = *ptr;
        newval.dbl = oldval.dbl + value;
    } while( ! __sync_bool_compare_and_swap((volatile uint64_t*)ptr, oldval.ui64, newval.ui64));

    return oldval.dbl;
}

// fine-grain locking through atomic sync var
static inline double atomic_fglock_add_dbl(double* ptr, double value) {
    static volatile uint64_t synchronized = 0;

    while( ! __sync_bool_compare_and_swap((volatile uint64_t*)&synchronized,0,1) )
        ; // busy waiting
    double oldval  = *ptr;
    *ptr          += value;
    synchronized   = 0;

    return oldval;
}

#endif // DOTPRODUCT_H_GUARD
