#ifndef LINKED_LIST_H_GUARD
#define LINKED_LIST_H_GUARD

#include "pthread_wrappers.h"

/**
 * A link in a linked list -- data is symbolized by an int
 */
typedef struct pthd_link_s {
    struct pthd_link_s *next;
    pthread_t      id;
} pthd_link_t;


/**
 * Linked List Abstract Data Type
 */
typedef struct pthd_linked_list_s {
    pthd_link_t   *head,
             *tail;
    size_t    size;
    pthd_link_t* (*new_link)     (pthread_t);
    void    (*destroy_link) (pthd_link_t*, pthread_t*);
    void    (*push_back)    (struct pthd_linked_list_s*, pthread_t);
    void    (*push_front)   (struct pthd_linked_list_s*, pthread_t);
    pthd_link_t* (*find)         (struct pthd_linked_list_s* list, pthd_link_t* start, pthread_t value);

} pthd_linked_list_t;

void pthd_list_init (struct pthd_linked_list_s*, 
                     pthd_link_t* (*new_link) (pthread_t), 
                     void (*destroy_link) (pthd_link_t*, pthread_t*),
                     void (*push_back)  (pthd_linked_list_t*,pthread_t), 
                     void (*push_front) (pthd_linked_list_t*,pthread_t),
                     pthd_link_t* (*find) (pthd_linked_list_t*,pthd_link_t*,const pthread_t));

pthd_link_t* pthd_link_init    (pthread_t);
void         pthd_link_destroy (pthd_link_t*,pthread_t*);
void         pthd_list_destroy (pthd_linked_list_t*);
void         pthd_push_back    (pthd_linked_list_t*,pthread_t);
void         pthd_push_front   (pthd_linked_list_t*,pthread_t);
pthd_link_t* pthd_find           (pthd_linked_list_t*,pthd_link_t*,const pthread_t);
void pthd_remove_after(pthd_linked_list_t* list, pthd_link_t* l, pthread_t* value);
void pthd_insert_after(pthd_linked_list_t* list, pthd_link_t* l, pthread_t  value);

//void process_list(pthd_linked_list_t*,void (*processor)(pthd_link_t*));

#endif // LINKED_LIST_H_GUARD
