#include "utils.h"
#include <stddef.h>

const double EPSILON = 0.0001;

void
seq_matvec(double* v, const double* A, const double* b, size_t m, size_t n)
{
    for (size_t i = 0; i < m; ++i) {
        double acc = 0.0;
        for (size_t j = 0; j < n; ++j) {
            acc += A[i*n+j] * b[j];
        }
        v[i] = acc;
    }
}


void
omp_matvec(double* v, const double* A, const double* b, size_t m, size_t n)
{
    #pragma omp parallel for default(none) shared (v,A,b,m,n)
    for (size_t i = 0; i < m; ++i) {
        double acc = 0.0;
        for (size_t j = 0; j < n; ++j) {
            acc += A[i*n+j] * b[j];
        }
        v[i] = acc;
    }
}

const size_t M = 10;
const size_t N = 10;

int
main(void)
{
    double *A = smalloc(M * N * sizeof(double)),
           *b = smalloc(    N * sizeof(double)),
           *v = scalloc(N,      sizeof(double));

    
    #pragma omp parallel default(none) shared(A,b,v)
    {
        #pragma omp for nowait
        for (size_t i = 0; i < M; ++i) 
            for (size_t j = 0; j < N; ++j) 
                A[i*N+j] = i + j*2.01;

        #pragma omp for
        for (size_t i = 0; i < M; ++i) 
            b[i] = 10.3 * i + 1;
    }

    omp_matvec(v,A,b,M,N);

    // Sequential run
    double *v_seq = scalloc(N,      sizeof(double));
    seq_matvec(v_seq,A,b,M,N);

int errors=0;
    for (size_t i = 0; i < M; ++i) {
        if (v_seq[i]-v[i] > EPSILON || v_seq[i]-v[i] < -EPSILON)
	{
            printf("v_seq[%lu] = %.4f != v[%lu] = %.4f\n", i, v_seq[i], i, v[i]);
		errors++;
	}
    }
if(errors!=0)
{
printf("%d errors found!\n", errors);
}
else
{
printf("No errors found!\n");
}

    sfree(v_seq);
    sfree(v);
    sfree(b);
    sfree(A);
}
