#include "utils.h"
#include "pthread_wrappers.h"

const uint64_t N_THREADS = 100;

typedef struct id_s {
    uint64_t           id;
    volatile uint64_t* token;
    pthread_mutex_t*   lock;
    pthread_cond_t*    cond;
} id_t;

void*
worker(void* arg)
{
    id_t* frame = (id_t*) arg;

    spthread_mutex_lock(frame->lock); // Begin critical section

    while (*(frame->token) != frame->id)
        pthread_cond_wait(frame->cond,frame->lock);
    printf("[%lu]\tI have the token!\n", frame->id);
    (*frame->token)++; // or alternatively *frame->token = frame->id + 1;

    spthread_mutex_unlock(frame->lock); // End critical section

    spthread_cond_broadcast(frame->cond); // wake up everyone -- only one will stay awake

    return NULL;
}

int
main(void)
{
    pthread_t       threads [N_THREADS];
    id_t            ids     [N_THREADS];
    pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
    pthread_cond_t  cond = PTHREAD_COND_INITIALIZER;
    volatile uint64_t token = 0;

    for (uint64_t i = 0; i < N_THREADS; ++i) {
        ids[i] = (id_t){ .id = i, .token = &token, .lock = &lock, .cond = &cond };
        spthread_create(&threads[i], NULL, worker, &ids[i]);
    }

    for (uint64_t i = 0; i < N_THREADS; ++i) 
        spthread_join(threads[i],NULL);

    return 0;
}
